% Persuasion dialog (08/06/2015)

:- op(100,  fy, not).
:- op(101, xfy, and).
:- op(102, xfx, then).
:- op(103, xfx, :).
:- op(105, xfx, precedes).
:- op(200, xfx, #).
:- discontiguous precedence_relation/2.
:- dynamic tmp/1.     % temporary clause
:- dynamic next/2.    % next legal speech act
:- dynamic ptype/1.   % precedence type (explicit, implicit, mixed)
:- dynamic verbose/1. % activate the output of speech acts

/*----------------------------------------------------------------+
| Persuasion dialog: the agent assumes both roles in the dialog   |
| (i.e., pro and con).                                            |
+----------------------------------------------------------------*/

% +Domain # +Literal: simulates a persuasion dialogue whose goal is
% to prove that Literal is justified by the Domain's knowledge.
% The Domain is the name of a file with the Domain knowledebase.

Domain # Literal :-
   load(Domain),
   flag(arg,_,1),
   retractall(next(start,_)),
   asserta(next(start,pro:claim(Literal))),
   nb_setval(narrative,[]),
   ignore(dialogue(start)),
   outcome(Literal,Outcome,Color),
   ansi_format([bold,fg(Color)],'~n~w',[Outcome]),
   format(': ~w~n', [Literal]), !.

% dialogue(+Act): simulates a dialogue between 'pro' and 'con'
% (where both roles are played by the same agent), where pro tries
% to defend a claim. To start the simulation, Act must be 'start'.

dialogue(Act) :-
   next(Act,NextAct),
   new(NextAct),
   say(NextAct),
   upd(NextAct),
   not(final(NextAct)),
   not(dialogue(NextAct)).

% new(+Agent:+Locution): is true if Locution is new in the
% narrative (only agent 'con' can repeat since locutions).

new(Agent:Locution) :-
   ( Agent = con, Locution = since(_)
   ; nb_getval(narrative,Narrative),
     not(memberchk(_:Locution,Narrative)) ).

% upd(+Agent:Locution): updates the current narrative with the
% speech act Agent:Locution.

upd(Agent:Locution) :-
   nb_getval(narrative,Narrative),
   nb_setval(narrative,[Agent:Locution|Narrative]).

% say(+Agent:+Locution): write the speech act Agent:Locution to the
% current output (if verbose mode is active).

say(Agent:Locution) :-
   (  verbose(active)
   -> flag(arg,N,N+1),
      format('<~w> ~w : ~w',[N,Agent,Locution]),
      prompt1(''), get0(_)
   ;  true ).

/*----------------------------------------------------------------+
| Protocol rules                                                  |
+----------------------------------------------------------------*/

% Transition rules (legal moves)
% - A and B are adversarial agents in the dialog
% - F is a presumed fact and C is its complement
% - R and S are defeasible rules

next(A:claim(F),B:why(F)    ) :- adv(A,B).
next(A:claim(F),B:agree(F)  ) :- adv(A,B).
next(A:why(F),  B:since(R)  ) :- adv(A,B), adduces(R,F).
next(A:why(F),  B:retract(F)) :- adv(A,B).
next(A:since(R),B:why(P)    ) :- adv(A,B), premise(P,R).
next(A:since(R),B:since(S)  ) :- adv(A,B), attacks(B,S,R).
next(A:since(R),B:agree(F)  ) :- adv(A,B), claim(R,F).

% Termination rules

final(_:agree(_)).
final(_:retract(_)).

% Outcome rules
% - Literal is the claimed fact
% - Outcome is the result of the dialog

outcome(Literal,Outcome,Color) :-
  nb_getval(narrative,Narrative),
  Narrative \= [],
  (  member(con:agree(Literal),Narrative)
  -> (Outcome,Color) = ('Accepted',blue)
  ;  (Outcome,Color) = ('Rejected',red) ).

% adv(?Agent,?Adversary): is true if Agent and Adversary have
% opposite roles in a soliloquy.

adv(pro,con).
adv(con,pro).

% neg(?Literal,?Complement): is true if Literal and Complement are
% complementary literals.

neg(not(Proposition),Proposition) :- !.
neg(Proposition,not(Proposition)).

% coherent(+C): is true if the conjunction C is coherent with the
% commitements of the agent in current state of the narrative.

coherent(A then C) :-
   nb_getval(narrative,Narrative),
   forall(in(L,C and A),not(member(_:retract(L),Narrative))),
   forall(in(L,C and A),not((neg(L,N),member(_:agree(N),Narrative)))),
   forall(in(L,A),not((member(_:why(L),Narrative),not(member(_:agree(L),Narrative))))),
   forall(in(L,A),not((neg(L,N),member(_:why(N),Narrative),not(member(_:retract(N),Narrative))))).

% adduces(+R,+F): is true if rule R adduces fact F and is coherent with
% the current state of the narrative.

adduces(R: A then C,C) :-
   rule(R: A then C),
   coherent(A then C).

% premise(-L,+R): is true if L is a premise in rule R

premise(Literal,_:Antecedent then _) :-
   in(Literal,Antecedent),
   neg(Literal,Complement),
   coherent(Complement then nil).

% attacks(+A,+R1,+R2): is true if agent A can use rule R1 to attack
% rule R2.

attacks(B,S: C then N,R:_ then F) :-
   neg(F,N),
   rule(S: C then N),
   coherent(C then N),
   defeats(B,S,R).

% claim(+Rule,-Claim)

claim(_:_ then C,C).

/*----------------------------------------------------------------+
| Defeating rules		                                  |
+----------------------------------------------------------------*/

defeats(pro,S,R) :- precedes(S,R), !.
defeats(con,S,R) :- precedes(S,R), !.
defeats(con,S,R) :- incomparable(S,R).

precedes(L1,L2) :-
   nb_getval(precedences,Precedences),
   memberchk(L1<L2,Precedences).

incomparable(S,R) :- not(precedes(S,R)), not(precedes(R,S)).

/*----------------------------------------------------------------+
| Explicit preference relation                                    |
+----------------------------------------------------------------*/

% precedence_relation(explicit,-Precedences): is true if Precedences
% is the set of explitit precedence rules synthesized from precedence
% rules declared in the knowledge base.

precedence_relation(explicit,Precedences) :-
   nb_getval(kb,Clauses),
   findall(L1<L2,member(precedes(L1,L2),Clauses),Explicit),
   transitive_closure(Explicit,Closure),
   filtered(Closure,Clauses,Precedences).

% transitive_closure(+R,-T): is true if the transitive closure of the
% acyclic binary relation R is T

transitive_closure(R,T) :-
   acyclic(R),
   ( setof(L1<L2,transition(L1,L2,R),T), ! ; T=[] ).

% acyclic(+Relation): is true if the Relation is acyclic

acyclic([]) :- !.
acyclic(Relation) :-
   select(L1<_,Relation,NewRelation),
   not(member(_<L1,Relation)),
   acyclic(NewRelation), !.

% transition(-L1,-L2,+R): is true if the transition L1<L2 can be
% inferred from the precedence rules in the relation R.

transition(L1,L2,R) :- member(L1<L2,R).
transition(L1,L2,R) :- member(L1<L,R), transition(L,L2,R).

% Elimiate precedences between rules which have no complementary
% consequents (since they are useless).

filtered(Closure,Clauses,Precedences) :-
   findall(Label1<Label2,
	   ( member(Label1<Label2,Closure),
	     member(Label1:then(_,Consequent1),Clauses),
	     member(Label2:then(_,Consequent2),Clauses),
	     neg(Consequent1,Consequent2) ),
	   Precedences).

/*----------------------------------------------------------------+
| Implicit precedence relation                                    |
+----------------------------------------------------------------*/

% precedence_relation(implicit,-Precedences): is true if Precedences is
% the set of implitit precedence rules synthesized from defeasible
% rules declared in the knowledge base.

precedence_relation(implicit,Precedences) :-
   nb_getval(kb,Clauses),
   findall(Label,member(Label:_,Clauses),Labels),
   findall(L1<L2,
	   ( member(L1,Labels),
             member(L2,Labels),
             L1 \= L2,
	     more_specific(L1,L2) ),
	   Implicit),
   transitive_closure(Implicit,Closure),
   filtered(Closure,Clauses,Precedences).

% more_specific(L1,L2): is true if: (1) L1 is not a presumption, (2) L1
% and L2 are conflicting rules, (3) the antecedent of L2 can be derived
% from the antecedent of L1, and (4) and the antecedent of L1 cannot be
% derived from the antecedent of L1.

more_specific(L1,L2) :-
   rule(L1: A1 then C1),
   rule(L2: A2 then C2),
   neg(C1,C2),
   der(A2,A1),
   not(der(A1,A2)), !.

% der(+C1,+C2): is true if conjunction C1 is defeasibly derivable from
% the conjunction C2.

der(C1,C2) :-
   setof(L,in(L,C2),Ls),
   forall(in(L,C1),ddt([L],Ls)).

% defeasible derivation tree

ddt([L|H],_) :- memberchk(L,H), !, fail. % avoid cyclic rules
ddt([L|_],Ls) :- memberchk(L,Ls), !.
ddt([L|H],Ls) :-
   rule(_: A then L), in(X,A),
   ddt([X,L|H],Ls).

/*----------------------------------------------------------------+
| Mixed precedence relation                                       |
+----------------------------------------------------------------*/

% precedence_relation(mixed,-Precedences): is true if
% Precedences is the set of mixed precedence rules synthesized
% from the integration of implicit and explicit precedence
% relations.

precedence_relation(mixed,Precedences) :-
   nb_getval(kb,Clauses),
   precedence_relation(explicit,Explicit),
   precedence_relation(implicit,Implicit),
   integrate(Explicit,Implicit,Mixed),
   transitive_closure(Mixed,Closure),
   filtered(Closure,Clauses,Precedences).

% Integrate explicit and implicit precedence relations
% (prioritizing explicit precedences).

integrate(Explicit,Implicit,Mixed) :-
   union(Explicit,Implicit,Union),
   purge(Union,Explicit,Mixed).

% Eliminate cycles caused by explcit precedences, by
% removing the weakest implicit precedences in the Relation.

purge(Relation,Explicit,Purged) :-
   reject(Relation,Explicit,Links), !,
   %delete(Relation,Links,NewRelation),
   subtract(Relation,Links,NewRelation),
   purge(NewRelation,Explicit,Purged).
purge(Relation,_,Relation).

% reject(+Relation,+Explicit,-Link): is true if Relation is cyclic and
% Link is the weakest link in a shortest cycle in this relation.

reject(Relation,Explicit,Links) :-
   setof([L1],L2^member(L1<L2,Explicit),Sources),
   shortest(Sources,Relation,Cycle),
   weakest(Cycle,Explicit,Links), !,
   Links \= [].

% shortest(+ListOfSources,+Relation,-Cycle): is true if Cycle is a
% shortest cycle in the Relation, starting from one of the labels in the
% ListOfSources. If Relation is acyclic, Cycle is an empty path.

shortest([],_,[]) :- !.
shortest([[Label|Labels]|_Paths],_Relation,[Label|Labels]) :-
   member(Label,Labels), !.
shortest([Path|Paths],Relation,Cycle) :-
   Path = [Label|_],
   findall([NextLabel|Path],member(Label<NextLabel,Relation),Successors),
   append(Paths,Successors,NewForest),
   shortest(NewForest,Relation,Cycle).

% weakest(+Cycle,+Explicit,-Links): is true if Links is the set
% of weakest links in the Cycle. The weakest links in the Cycle is
% all implicit links (derived from an implicit precedence) that
% immediately precedes an explict link (derived from an explicit
% precedence).

weakest(Cycle,Explicit,Links) :-
   findall(L1<L2,
	   ( append(_,[L2,L1|_],Cycle),	  % find link that
	     not(member(L1<L2,Explicit)), % is an implicit link and
	     memberchk(L2<_,Explicit) ),  % precedes an explicit link
	   Links).

/*----------------------------------------------------------------+
| Knowledge base creation                                         |
+----------------------------------------------------------------*/

% load(+Domain,-Clauses): read Clauses from a file defining the
% knowledgebase for specific Domain and compute the precedence
% relation to be used.

load(Domain) :-
   atomic_list_concat(['domains/',Domain,'.pl'],FileName),
   getkb(FileName,Clauses),
   unflawed(Clauses),
   nb_setval(kb,Clauses),
   ignore(not(gratuitous(Clauses))),
   ptype(Type),
   (  precedence_relation(Type,Precedences)
   -> nb_setval(precedences,Precedences)
   ;  ansi_format([bold,fg(black)],'~nError: ',[]),
      format('cyclic precedence relation~n~n'), fail ).

getkb(FileName,Clauses) :-
   see(FileName),
   repeat,
      read(Line),
      assertz(tmp(Line)),
   Line = end_of_file, !,
   seen,
   retract(tmp(end_of_file)),
   findall(Clause,tmp(Clause),Clauses),
   retractall(tmp(_)).

% unflawed(SetOfClauses): is true if all defeasible rules in
% SetOfClauses are consistent rules.

unflawed([]) :- !.
unflawed([precedes(_,_)|SetOfClauses]) :- !,
   unflawed(SetOfClauses).
unflawed([_:then(A,C)|SetOfClauses]) :-
   consistent(then(A,C)), !,
   unflawed(SetOfClauses).
unflawed([L:then(A,C)|_]) :-
   ansi_format([bold,fg(black)],'~nError: ',[]),
   format('inconsistent rule (~w)~n~n',[L:then(A,C)]),
   !, fail.

% consistent(+then(Antecedent,Consequent)): is true if
% 'Antecedent and Consequent' is consistent.

consistent(then(Antecedent,Consequent)) :-
   findall(Premise,in(Premise,Antecedent),Premises),
   not( ( member(Proposition1,[Consequent|Premises]),
          member(Proposition2,[Consequent|Premises]),
          neg(Proposition1,Proposition2) ) ), !.

% gratuitous(+SetOfClauses): is tru if the SetOFClauses has a rule
% with a gratuitous consequent (e.g., 'a then a', or 'p then q' and
% 'q then p')

gratuitous(Clauses) :-
   findall(R,( member(R,Clauses),gratuitous(R,Clauses)),G), G\=[],
   ansi_format([bold,fg(black)],'~nWarning: ',[]),
   format('gratuitous consequents (~w)~n~n',[G]).

gratuitous(_:then(A,C),_) :- in(C,A), !.
gratuitous(_:then(A1,C1),Clauses) :-
   member(_:then(A2,C2),Clauses),
   in(L1,A1), der(L1,C2),
   in(L2,A2), der(L2,C1), !.

/*----------------------------------------------------------------+
| Auxiliary predicates					          |
+----------------------------------------------------------------*/

% rule(?Rule): is true if Rule is in the knowledge base.

rule(Rule) :- nb_getval(kb,Clauses), member(Rule,Clauses).

% in(?L,+C): is true if L is a literal in conjunction C.

in(L,L) :- L \= true, L \= and(_,_).
in(L,and(L,_)).
in(L,and(_,C)) :- in(L,C).

% sset(+A,+B): is true if each literal in conjunction A is also a
% literal in conjunction B.

sset(A,B) :- forall(in(X,A),in(X,B)).

% Activate/Deactivate verbose mode

:- retractall(verbose(_)), assert(verbose(active)).

verbose :- retract(verbose(active)), assert(verbose(inactive)), !.
verbose :- retract(verbose(inactive)), assert(verbose(active)), !.

% Set precedence type (explicit, implicit or mixed)

:- retractall(ptype(_)), assert(ptype(mixed)).

explicit :- retractall(ptype(_)), assert(ptype(explicit)).
implicit :- retractall(ptype(_)), assert(ptype(implicit)).
mixed    :- retractall(ptype(_)), assert(ptype(mixed)).

% Show all precedence relations for a Domain

precedence_relations(Domain) :-
   load(Domain),
   precedence_relation(explicit,Explicit),
   precedence_relation(implicit,Implicit),
   precedence_relation(mixed,Mixed),
   format('Explicit: ~w~n',[Explicit]),
   format('Implicit: ~w~n',[Implicit]),
   format('Mixed...: ~w~n',[Mixed]).












